"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
var template_1 = __importDefault(require("@babel/template"));
var t = __importStar(require("@babel/types"));
var traverse_1 = __importDefault(require("@babel/traverse"));
var utils_1 = require("@compiled/utils");
var css_1 = require("@compiled/css");
var is_prop_valid_1 = __importDefault(require("@emotion/is-prop-valid"));
var ast_1 = require("./ast");
/**
 * Hoists a sheet to the top of the module if its not already there.
 * Returns the referencing identifier.
 *
 * @param sheet Stylesheet
 * @param meta Plugin metadata
 */
var hoistSheet = function (sheet, meta) {
    if (meta.state.sheets[sheet]) {
        return meta.state.sheets[sheet];
    }
    var sheetIdentifier = meta.parentPath.scope.generateUidIdentifier('');
    var parent = meta.parentPath.findParent(function (path) { return path.isProgram(); }).get('body');
    var path = parent.filter(function (path) { return !path.isImportDeclaration(); })[0];
    path.insertBefore(t.variableDeclaration('const', [t.variableDeclarator(sheetIdentifier, t.stringLiteral(sheet))]));
    meta.state.sheets[sheet] = sheetIdentifier;
    return sheetIdentifier;
};
/**
 * Will build up the CSS variables prop to be placed as inline styles.
 *
 * @param variables CSS variables that will be placed in the AST
 * @param transform Transform function that can be used to change the CSS variable expression
 */
var buildCssVariablesProp = function (variables, transform) {
    if (transform === void 0) { transform = function (expression) { return expression; }; }
    return utils_1.unique(
    // Make sure all defined CSS variables are unique
    variables, 
    // We consider their uniqueness based on their name
    function (item) { return item.name; }).map(function (variable) {
        // Map them into object properties.
        return t.objectProperty(t.stringLiteral(variable.name), 
        // Allow callers to transform the expression if needed,
        // for example the styled API strips away the arrow function.
        transform(variable.expression));
    });
};
/**
 * Builds up the inline style prop value for a Styled Component.
 *
 * @param variables CSS variables that will be placed in the AST
 * @param transform Transform callback function that can be used to change the CSS variable expression
 */
var styledStyleProp = function (variables, transform) {
    var props = [t.spreadElement(t.identifier('style'))];
    return t.objectExpression(props.concat(buildCssVariablesProp(variables, transform)));
};
/**
 * Returns a tag string in the form of an identifier or string literal.
 *
 * A type of InBuiltComponent will return a string literal,
 * otherwise an identifier string will be returned.
 *
 * @param tag Made of name and type.
 */
var buildComponentTag = function (_a) {
    var name = _a.name, type = _a.type;
    return type === 'InBuiltComponent' ? "\"" + name + "\"" : name;
};
/**
 * Traverses an arrow function and then finally return the arrow function body node.
 *
 * @param node Array function node
 * @param nestedVisitor Visitor callback function
 */
var traverseStyledArrowFunctionExpression = function (node, nestedVisitor) {
    traverse_1.default(node, nestedVisitor);
    return ast_1.pickFunctionBody(node);
};
/**
 * Traverses a binary expression looking for any arrow functions,
 * calls back with each arrow function node into the passed in `nestedVisitor`,
 * and then finally replaces each found arrow function node with its body.
 *
 * @param node Binary expression node
 * @param nestedVisitor Visitor callback function
 */
var traverseStyledBinaryExpression = function (node, nestedVisitor) {
    traverse_1.default(node, {
        noScope: true,
        ArrowFunctionExpression: function (path) {
            path.traverse(nestedVisitor);
            path.replaceWith(ast_1.pickFunctionBody(path.node));
            path.stop();
        },
    });
    return node;
};
/**
 * Will return a generated AST for a Styled Component.
 *
 * @param opts Template options
 * @param meta Plugin metadata
 */
var styledTemplate = function (opts, meta) {
    var nonceAttribute = meta.state.opts.nonce ? "nonce={" + meta.state.opts.nonce + "}" : '';
    var propsToDestructure = [];
    var styleProp = opts.variables.length
        ? styledStyleProp(opts.variables, function (node) {
            var nestedArrowFunctionExpressionVisitor = {
                noScope: true,
                MemberExpression: function (path) {
                    if (t.isIdentifier(path.node.object) && path.node.object.name === 'props') {
                        var propertyAccessName = path.node.property;
                        if (is_prop_valid_1.default(propertyAccessName.name)) {
                            return;
                        }
                        if (!propsToDestructure.includes(propertyAccessName.name)) {
                            propsToDestructure.push(propertyAccessName.name);
                        }
                        path.replaceWith(propertyAccessName);
                    }
                },
            };
            if (t.isArrowFunctionExpression(node)) {
                return traverseStyledArrowFunctionExpression(node, nestedArrowFunctionExpressionVisitor);
            }
            if (t.isBinaryExpression(node)) {
                return traverseStyledBinaryExpression(node, nestedArrowFunctionExpressionVisitor);
            }
            return node;
        })
        : t.identifier('style');
    return template_1.default("\n  React.forwardRef(({\n    as: C = " + buildComponentTag(opts.tag) + ",\n    style,\n    " + propsToDestructure.map(function (prop) { return prop + ','; }).join('') + "\n    ...props\n  }, ref) => (\n    <CC>\n      <CS " + nonceAttribute + ">{%%cssNode%%}</CS>\n      <C\n        {...props}\n        style={%%styleProp%%}\n        ref={ref}\n        className={ax([" + opts.classNames
        .map(function (className) { return "\"" + className + "\""; })
        .join(',') + ", props.className])}\n      />\n    </CC>\n  ));\n", {
        plugins: ['jsx'],
    })({
        styleProp: styleProp,
        cssNode: t.arrayExpression(opts.sheets.map(function (sheet) { return hoistSheet(sheet, meta); })),
    });
};
/**
 * Will return a generated AST for a Compiled Component.
 * This is primarily used for CSS prop and ClassNames apis.
 *
 * @param node Originating node
 * @param sheets Stylesheets
 * @param meta Metadata
 */
var compiledTemplate = function (node, sheets, meta) {
    var nonceAttribute = meta.state.opts.nonce ? "nonce={" + meta.state.opts.nonce + "}" : '';
    return template_1.default("\n  <CC>\n    <CS " + nonceAttribute + ">{%%cssNode%%}</CS>\n    {%%jsxNode%%}\n  </CC>\n  ", {
        plugins: ['jsx'],
    })({
        jsxNode: node,
        cssNode: t.arrayExpression(sheets.map(function (sheet) { return hoistSheet(sheet, meta); })),
    });
};
/**
 * Will join two expressions together,
 * Looks like `left + ' ' + right`.
 *
 * @param left Any node on the left
 * @param right Any node on the right
 * @param spacer Optional spacer node to place between the left and right node. Defaults to a space string.
 */
exports.joinExpressions = function (left, right, spacer) {
    if (spacer === void 0) { spacer = t.stringLiteral(' '); }
    return t.binaryExpression('+', left, spacer ? t.binaryExpression('+', spacer, right) : right);
};
/**
 * Will conditionally join two expressions together depending on the right expression.
 * Looks like: `left + right ? ' ' + right : ''`
 */
exports.conditionallyJoinExpressions = function (left, right) {
    return t.binaryExpression('+', left, t.conditionalExpression(right, t.binaryExpression('+', t.stringLiteral(' '), right), t.stringLiteral('')));
};
/**
 * Returns a Styled Component AST.
 *
 * @param tag Styled tag either an inbuilt or user define
 * @param cssOutput CSS and variables to place onto the component
 * @param meta Plugin metadata
 */
exports.buildStyledComponent = function (tag, cssOutput, meta) {
    var _a = css_1.transformCss(cssOutput.css), classNames = _a.classNames, sheets = _a.sheets;
    return styledTemplate({
        classNames: classNames,
        tag: tag,
        sheets: sheets,
        variables: cssOutput.variables,
    }, meta);
};
/**
 * Wrapper to make defining import specifiers easier.
 * If `localName` is defined it will rename the import to it,
 * e.g: `name as localName`.
 *
 * @param name import name
 * @param localName local name
 */
exports.importSpecifier = function (name, localName) {
    return t.importSpecifier(t.identifier(name), t.identifier(localName || name));
};
/**
 * Returns the actual value of a jsx value.
 *
 * @param node
 */
exports.getPropValue = function (node) {
    var value = t.isJSXExpressionContainer(node) ? node.expression : node;
    if (t.isJSXEmptyExpression(value)) {
        throw new Error('Empty expression not supported.');
    }
    return value;
};
/**
 * Returns a Compiled Component AST.
 *
 * @param node Originating node
 * @param cssOutput CSS and variables to place onto the component
 * @param meta Plugin metadata
 */
exports.buildCompiledComponent = function (node, cssOutput, meta) {
    var _a = css_1.transformCss(cssOutput.css), sheets = _a.sheets, classNames = _a.classNames;
    var classNameProp = node.openingElement.attributes.find(function (prop) {
        return t.isJSXAttribute(prop) && prop.name.name === 'className';
    });
    if (classNameProp && classNameProp.value) {
        // If there is a class name prop statically defined we want to concatenate it with
        // the class name we're going to put on it.
        var classNameExpression = exports.getPropValue(classNameProp.value);
        var values = classNames
            .map(function (className) { return t.stringLiteral(className); })
            .concat(classNameExpression);
        classNameProp.value = t.jsxExpressionContainer(t.callExpression(t.identifier('ax'), [t.arrayExpression(values)]));
    }
    else {
        // No class name - just push our own one.
        node.openingElement.attributes.push(t.jsxAttribute(t.jsxIdentifier('className'), t.jsxExpressionContainer(t.callExpression(t.identifier('ax'), [
            t.arrayExpression(classNames.map(function (name) { return t.stringLiteral(name); })),
        ]))));
    }
    if (cssOutput.variables.length) {
        // If there is dynamic CSS in use we have work to do.
        var stylePropIndex_1 = -1;
        // Find the style prop on the opening JSX element.
        var styleProp = node.openingElement.attributes.find(function (prop, index) {
            if (t.isJSXAttribute(prop) && prop.name.name === 'style') {
                stylePropIndex_1 = index;
                return true;
            }
            return false;
        });
        var dynamicStyleProperties_1 = buildCssVariablesProp(cssOutput.variables);
        if (styleProp) {
            // Remove the pre-existing style prop - we're going to redefine it soon.
            node.openingElement.attributes.splice(stylePropIndex_1, 1);
            if (styleProp.value &&
                t.isJSXExpressionContainer(styleProp.value) &&
                !t.isJSXEmptyExpression(styleProp.value.expression)) {
                // If it's not an object we just spread the expression into the object
                if (!t.isObjectExpression(styleProp.value.expression)) {
                    dynamicStyleProperties_1.splice(0, 0, t.spreadElement(styleProp.value.expression));
                }
                else {
                    // Else it's an object! So we want to place each property into the object
                    styleProp.value.expression.properties.forEach(function (prop, index) {
                        if (t.isObjectMethod(prop)) {
                            return;
                        }
                        // We want to keep the order that they were defined in.
                        // So we're using index here to do just that.
                        dynamicStyleProperties_1.splice(index, 0, prop);
                    });
                }
            }
        }
        // Finally add the new style prop back to the opening JSX element.
        node.openingElement.attributes.push(t.jsxAttribute(t.jsxIdentifier('style'), t.jsxExpressionContainer(t.objectExpression(dynamicStyleProperties_1))));
    }
    return compiledTemplate(node, sheets, meta);
};
//# sourceMappingURL=ast-builders.js.map